#!/usr/bin/perl
#
# $Id: run_acspo.pl 1079 2017-01-20 21:20:39Z jimd $
#
# Top-level perl script for CSPP_ACSPO package.
# Uses only Perl modules that are part of the v5.006 standard install.
#
# A set of user-defined functions are appended to this main script;
# they are listed in alphabetical order.
#
# The following notes were made for wrapping ACSPO v2.31, but I believe that they
# still apply, in the main, to ACSPO v2.40 ...
#
# Pre-requisites:
# ACSPO distribution in directory $CSPP_ACSPO_HOME/$CSPP_ACSPO_REVNAME
# In this script $HOME=$CSPP_ACSPO_HOME/$CSPP_ACSPO_REVNAME for short.
# Various PCFs (paths & configuration files) are in $HOME/run/{avhrr,modis,viirs}
# directories to serve as templates. There is just one file that is specifically
# named in main_module.f90 and that is ./Output_Data_Structure.txt.
#
# $ grep Output_Data_Structure.txt acspo-?/main_module.f90
# acspo-a/main_module.f90:character(len=*),parameter:: path = './Output_Data_Structure.txt'
# acspo-m/main_module.f90:character(len=*),parameter:: path = './Output_Data_Structure.txt'
# acspo-v/main_module.f90:character(len=*),parameter:: path = './Output_Data_Structure.txt'
#
# This will need to be made verbatim in the runtime work directory. Within
# here the ouput directory is specified, as are the remaining PCF files, namely
#     sensorConfigFile.txt
#         input_dir=input/
#         input_files_list=1bFilesList.txt
#     simDataConfigFile.txt
#     Cloud_Mask_Config.txt
# 
# Additionally, for avhrr processing ('n15', 'n16', 'n17', 'n18', 'n19', 'metopA', 'metopB')
# $ grep aerConfigFile acspo-?/*.f90
# acspo-a/aerosol.f90:    CHARACTER (LEN=*), PARAMETER :: CONFIG_FILE_NAME = 'aerConfigFile.txt'
#
#
# I do not want to fiddle (excpet where absolutely necessary) with any files in
# the delivery directories!
#
#
# START-HELP (N.B. This tag is used in printing help, do not touch!)
#
#  Options:
#  --version             show program's version number and exit
#  ?, -h, --help         show this help message and exit
#  -n, --nothing         make the PCF files, but do not execute.
#                        [Does the ancillary data preparation and, where
#                        applicable, destriping, just not the ACSPO run].
#
#  Mandatory Arguments:
#    At a minimum these arguments must be specified [i.e. no default].
#
#    -i INPUT_FILES, --input_files=INPUT_FILES
#                        Path to input files. May be directory name, or
#                        directory/filemask (use * and ?, not regexp).
#                        Wildcarded filemasks MUST be "quoted" (or OS will expand).
#                        e.g. -i /dataDir/"{GMODO,SVM{05,07,10,12,15,16}}*t18*.h5"
#                        OR   -i /dataDir/"a1.14219.*.{geo,1000m}.hdf" 
#                        (BUT -i /dataDir  really is the simplest way).
#                        N.B. To use GMTCO for geolocation add switch --xfiles
#    -s MISSION, --sat=MISSION
#                        The mission satellite data for ACSPO algorithm.
#                        Possible values are: 'aqua', 'terra', 'npp', 'n15'
#                        'n16', 'n17', 'n18', 'n19', 'metopA', 'metopB'.
#    -b BIAS_DIR, --bias=BIAS_DIR
#                        The directory in which bias files will be stored after
#                        the run. Also the directory from which bias files will
#                        be fetched, if they exist, and used for current job.
#
#  Extra Options:
#    These options may be used to customize behaviour of this program.
#
#    -w WORK_DIR, --work=WORK_DIR
#                        The directory in which all activity will occur,
#                        this is where output files will be put.
#                        [default: current directory].
#    -d DYNAMIC_ANC, --dynanc=DYNAMIC_ANC
#                        Path to root of dynamic ancillary data tree;
#                        beneath here ancillary data are organized as
#                        per the SSEC remote ancillary server for CSPP,
#                        http://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/,
#                        i.e. %Y_%m_%d_%j where %Y is four-digit year, %m is
#                        two-digit month, %d is two-digit day of month and
#                        %j is three-digit day of year.
#                        [$CSPP_ACSPO_HOME/data/dynamic_ancillary]
#    -r CMC_RES, --ghrsst=CMC_RES
#                        Resolution of CMC GHRSST ancillary data files to use
#                        [default: 0.1 (deg) from 1-Jan-2017, 0.2 before]
#    -v, --verbose       Each occurrence increases verbosity 1 level from
#                        [default: ERROR only] -v=WARNING -vv=INFO -vvv=DEBUG
#    -c, --clean         Deletes temporary working directory under WORK_DIR
#                        at process completion.
#                        [default: --clean; --noclean is the contrary option].
#    -a, --aggregate     Aggregates npp GM{OD,TC}O/SVM granules into swaths of up
#                        to 30 minutes prior to processing.
#                        [default: --agg; --noagg is the contrary option].
#    -q, --quarantine    Processing machine is not connected to the web for the
#                        purpose of fetching required dynamic ancillary data.
#                        [default: --noquarantine].
#    -x, --xfiles        Use NOAA/CLASS filenames for avhrr L1B files or
#                        NASA/LANCE for modis geo and radiance HDF files.
#                        If applied to npp, means use GMTCO for geolocation.
#                        [default: --noxfiles, so Direct Broadcast filenames
#                        if applied to avhrr or modis processing, or use GMODO
#                        files for geolocation if applied to npp processing].
#
# END-HELP (N.B. This tag is used in printing help, do not touch!)
#
#
# UN-DOCUMENTED EXTRA !  Do not expect this to be too robust; it is a hack so
# that I can make a 10 minute granule to precisely match the size of the
# delivery test data, that is all.
#
#    -g, --granule       Force a 10 minute granule for npp. This is done in
#                        the control file for viirsAggregator by rounding the
#                        start time up to, and the stop time down to, a 10
#                        minute boundary.
#                        [default: --nogranule].
#
# The following option is inactive; %options{processors} defaults to 1 and is
# never interrogated. Leave here in case I ever add multicore functionality.
#
#    -p PROCESSORS, --processors=PROCESSORS
#                        Number of cpus to use for ACSPO processing.
#
# NOTES-TO-SELF:
# -------------
#   $!=N to change exit code on die() to N; default is 255
#
#   --clean removes the temporary directory tree. If the user selects --noclean
#   the temporary directory is NOT removed and all ACSPO product files are 
#   available in the temporary directory structure. This option might be useful
#   for debugging and is necessary if the user wants access to product files
#   generated prior to the output netCDF files that are copied up to WORK_DIR.
#   It is then user responsibility to dispense with unwanted directories/files.
#
#
# This is a valid way to run a single S-NPP granule ...
#
#   ./run_acspo.pl -s npp \
#     -i /dataDir/"{GMODO,SVM{05,07,10,12,15,16}}_npp_d20131120_t0656260_*_cspp_dev.h5"
#
# The quotes are to prevent OS expanding wildcards/tokens on the commandline.
#
# For Metop-A files with NOAA filenaming convention, something like this ...
#
#   ./run_acspo.pl -x -s metopA \
#     -i /dataDir/NSS.FRAC.M2.D13329.S0346.E0529.B3684546.SV
#
# ... will run a single overpass or orbit file.
#
# But it is a whole lot simpler to put your data into a directory and to
# process it all with something like (a DB format NOAA-18 example now) ...
#
#   ./run_acspo.pl -s n18 -i /dataDir -d /ancDir -b /biasDir -w /outDir
#
#
# There is some fat in here, espcially when it comes to handling and checking
# dates and times. Some is because I can only use modules in the standard
# install for 5.006, some because I have requisitioned pieces old code of mine
# before I knew any better, and some because I still do not know any better.


# Pragmata
# ````````
use 5.006;
use strict;
use warnings;


# Load Perl modules - standard modules only
# `````````````````````````````````````````
use File::Spec;
use File::Copy;
use Data::Dumper;
use Getopt::Long;
use File::Basename;
use Text::ParseWords;
use Encode qw(decode);
use POSIX qw(ceil floor);
use Cwd qw(abs_path getcwd);
use Time::gmtime qw(:FIELDS);
use List::Util qw(sum min max);
use File::Path qw(mkpath rmtree);
use Time::Local qw(timegm_nocheck);
use File::Temp qw(tempfile tempdir);
use Scalar::Util qw(looks_like_number);
use I18N::Langinfo qw(langinfo CODESET);

# Global variables are all UPPERCASE throughout
# `````````````````````````````````````````````
our $AGG_WINDOW_MINUTES = 30; # 30-minute aggregation window for npp
our $MAIN = basename($0);
defined $ENV{CSPP_ACSPO_HOME} or die "$MAIN\:: CSPP_ACSPO_HOME not set.\nAbort";
defined $ENV{CSPP_ACSPO_REVNAME} or die "$MAIN\:: CSPP_ACSPO_REVNAME not set.\nAbort";
defined $ENV{CSPP_ACSPO_REVDATE} or die "$MAIN\:: CSPP_ACSPO_REVDATE not set.\nAbort";
our $HOME = File::Spec -> catfile($ENV{CSPP_ACSPO_HOME}, $ENV{CSPP_ACSPO_REVNAME});
(-d $HOME) or die "$MAIN\:: CSPP_ACSPO subdirectory $HOME not found.\nAbort";


# $H5DUMP points to supplied static-compiled h5dump (version 1.8.11)
# ``````````````````````````````````````````````````````````````````
our $H5DUMP = File::Spec -> catfile($ENV{'CSPP_ACSPO_HOME'}, 'bin', 'h5dump');
(-e $H5DUMP) or die "$MAIN\:: h5dump $H5DUMP not found.\nAbort";
(-x $H5DUMP) or die "$MAIN\:: h5dump $H5DUMP not executable.\nAbort";


# $HDP points to supplied static-compiled hdp (HDF Version 4.2 Release 10)
# ````````````````````````````````````````````````````````````````````````
our $HDP = File::Spec -> catfile($ENV{'CSPP_ACSPO_HOME'}, 'bin', 'hdp');
(-e $HDP) or die "$MAIN\:: hdp $HDP not found.\nAbort";
(-x $HDP) or die "$MAIN\:: hdp $HDP not executable.\nAbort";


# $H5REPACK points to supplied static-compiled h5dump (version 1.8.11)
# ````````````````````````````````````````````````````````````````````
our $H5REPACK = File::Spec -> catfile($ENV{'CSPP_ACSPO_HOME'}, 'bin', 'h5repack');
(-e $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not found.\nAbort";
(-x $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not executable.\nAbort";


# $H5NAGG points to supplied static-compiled nagg
# ```````````````````````````````````````````````
# But you can point to your own nagg if you wish. Ours is Version 1.5.1
our $H5NAGG = File::Spec -> catfile($ENV{'CSPP_ACSPO_HOME'}, 'bin', 'nagg');
(-e $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not found.\nAbort";
(-x $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not executable.\nAbort";


# $VIIRSAGG points to NOAA-supplied static-compiled viirsAggregator
# `````````````````````````````````````````````````````````````````
our $VIIRSAGG = File::Spec -> catfile($HOME, 'bin', 'viirsAggregator');
(-e $VIIRSAGG) or die "$MAIN\:: aggregator $VIIRSAGG not found.\nAbort";
(-x $VIIRSAGG) or die "$MAIN\:: aggregator $VIIRSAGG not executable.\nAbort";


# $GRIB2HDF points to NOAA-supplied static-compiled convert_grib_hdf 
# ``````````````````````````````````````````````````````````````````
our $GRIB2HDF = File::Spec -> catfile($HOME, 'bin', 'convert_grib_hdf');
(-e $GRIB2HDF) or die "$MAIN\:: grib2 converter $GRIB2HDF not found.\nAbort";
(-x $GRIB2HDF) or die "$MAIN\:: grib2 converter $GRIB2HDF not executable.\nAbort";


# Get the command line options, display help or revision if asked
# ```````````````````````````````````````````````````````````````
my $revision = '$Id: run_acspo.pl 1079 2017-01-20 21:20:39Z jimd $';
my $pwd = getcwd();
my %options = get_commandline($revision, $pwd);
($options{verbose} >= 2) and pretty_dump('%options', %options);


# At present we are saying that lftp is a system requirement
# ``````````````````````````````````````````````````````````
our $LFTP = 'lftp';
($options{quarantine}) or do {
  system("$LFTP --version 1>/dev/null 2>/dev/null") == 0
    or die "$MAIN\:: Required utility lftp not found\nAbort"; };


# Define a mapping for links from source -> target
# ````````````````````````````````````````````````
# The %source will be copied from the run directory of the specified sensor,
# (i.e. --sat=MISSION (from command line) now stored in $options{sat}) to
# the matching %target filename in the temporary working directory. Now these
# files are as supplied in the delivery - so we might need to change a few
# things to match the CSPP_ACSPO installation (like the location of static
# ancillary data)
# A WORD ABOUT HARD-CODED HASHES OF CONFIGUATION-LIKE KEY:VALUE PAIRS:
# In IMAPP projects (like IDEA-I) I have used a configuration file, specified
# on command line, to set what might be configuration-type values. In this CSPP
# project I want to avoid a configuration file and offer the user less choice.
# It turns out that at the command/configuration file level in ACSPO there is
# a great deal of flexibility. Apart from Output_Data_Structure.txt, which is
# mentioned explicitly in main_module.f90 of each sensor, there is full
# flexibility in naming files. BUT it makes sense to follow what has been
# done in the delivery examples; to make it easier for both me and the code
# originators to understand. You are going to see a bunch of hashes near the
# top which handle variable name mappings to specific values.  The deeper
# levels of the script can then be kept free of hard-coded string values.
# N.B. As at 17-Dec-2014, these files are actually named the same; i.e. the
# template in %source is the same as actual file to use in %target.  This
# used NOT to be the case - a choice of different templates was offered,
# e.g. Output_Data_Structure_GDS2.txt & Output_Data_Structure_Legacy.txt.
# BUT I am keeping the previous logic of a filename mapping anyway.
# N.B. aerConfigFile.txt is used ONLY in AVHRR processing.
my %sources = (pcf => 'Output_Data_Structure.txt',
               sen => 'sensorConfigFile.txt',
               sim => 'simDataConfigFile.txt',
               cld => 'Cloud_Mask_Config.txt',
               sst => 'SST_Regression_Config.txt',
               aer => 'aerConfigFile.txt');
my %targets = (pcf => 'Output_Data_Structure.txt',
               sen => 'sensorConfigFile.txt',
               sim => 'simDataConfigFile.txt',
               cld => 'Cloud_Mask_Config.txt',
               sst => 'SST_Regression_Config.txt',
               aer => 'aerConfigFile.txt');


# These are used for fileglobbing when INPUTFILES is a directory name
# ```````````````````````````````````````````````````````````````````
# Change to suit your site; for more control edit function validate_options(),
# or provide "<dirname>/<my_filemask>" on the command line as INPUTFILES; the
# "quotes" are important to prevent the OS expanding wildcard tokens. --xfiles
# is used on command line to signify that we are using non-DB names for input
# data files - did not matter for npp/viirs in previous version, but now I use
# --xfiles to flag use of GMTCO for geolocation instead of GMODO files. I am
# going to support TWO modis DB filenaming conventions 'a1.*.{geo,1000m}.hdf' &
# 'A*.{GEO,L1B_LAC}' (here using aqua as example). So for normal DB operation
# I have implemented a test for the 'a1.*.{geo,1000m}.hdf' form, and a switch
# to the 'A*.{GEO,L1B_LAC}' form if no files are found; see validate_options().
my %satdir = (aqua => 'modis',
              terra => 'modis',
              npp => 'viirs',
              n15 => 'avhrr',
              n16 => 'avhrr',
              n17 => 'avhrr',
              n18 => 'avhrr',
              n19 => 'avhrr',
              metopA => 'avhrr',
              metopB => 'avhrr');
my %filemask = (aqua => 'a1.*.{geo,1000m}.hdf',
                terra => 't1.*.{geo,1000m}.hdf',
                npp => '{GMODO,SVM{05,07,10,12,15,16}}*.h5',
                n15 => 'hrpt_noaa15*.l1b',
                n16 => 'hrpt_noaa16*.l1b',
                n17 => 'hrpt_noaa17*.l1b',
                n18 => 'hrpt_noaa18*.l1b',
                n19 => 'hrpt_noaa19*.l1b',
                metopA => 'hrpt_M02*.l1b',
                metopB => 'hrpt_M01*.l1b');
($options{xfiles}) and do {
  %filemask = (aqua => 'MYD0{21KM,3}*.hdf',
               terra => 'MOD0{21KM,3}*.hdf',
               npp => '{GMTCO,SVM{05,07,10,12,15,16}}*.h5',
               n15 => '???.????.NK.D?????.S????.E????.B*',
               n16 => '???.????.NL.D?????.S????.E????.B*',
               n17 => '???.????.NM.D?????.S????.E????.B*',
               n18 => '???.????.NN.D?????.S????.E????.B*',
               n19 => '???.????.NP.D?????.S????.E????.B*',
               metopA => '???.????.M2.D?????.S????.E????.B*',
               metopB => '???.????.M1.D?????.S????.E????.B*'); };
($options{verbose} == 3) and pretty_dump('%filemask', %filemask);


# Parse and validate command line %options & make temporary working directory
# ```````````````````````````````````````````````````````````````````````````
my $sat = $options{sat};
%options = validate_options($filemask{$sat}, $satdir{$sat}, %options);
my $tmpdir = tempdir('TMP_XXXXXX', DIR => $options{work},
  CLEANUP => $options{clean});
chdir($tmpdir) or die "$MAIN\:: Cannot chdir to temp dir $tmpdir\nAbort";
($options{verbose} >= 2) and print "chdir to working directory $tmpdir\n";


# Copy any bias files into place from BIAS_DIR
# `````````````````````````````````````````````
my %histmask = (aqua => '*_HISTOGRAM_AM.dat',
                terra => '*_HISTOGRAM_TM.dat',
                npp => '*_HISTOGRAM_NPP.dat',
                n15 => '*_HISTOGRAM_15_F.dat',
                n16 => '*_HISTOGRAM_16_F.dat',
                n17 => '*_HISTOGRAM_17_F.dat',
                n18 => '*_HISTOGRAM_18_F.dat',
                n19 => '*_HISTOGRAM_19_F.dat',
                metopA => '*_HISTOGRAM_2_F.dat',
                metopB => '*_HISTOGRAM_1_F.dat');
my $globber = File::Spec -> catfile($options{bias}, $histmask{$sat});
foreach (sort(glob($globber))) {
  copy($_, '.') or do {
    ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!"; };
}


# Preprocess and stage data for ACSPO system compatibility
# ````````````````````````````````````````````````````````
%options = cspp_preprocess(%options);
($options{verbose} == 3) and pretty_dump('%options', %options);
my %dynanc_dates = acspo_dynanc_dates(%options);
($options{verbose} == 3) and pretty_dump('%dynanc_dates', %dynanc_dates);


# If $options{ghrsst} eq 'by_date' it needs to be reset by input file date
# ````````````````````````````````````````````````````````````````````````
# This is a bit of a nuisance because we can have multiple input files.
# For maximum generality I would need to re-order the pre-processing steps
# in a pretty major way, all for the very unlikely scenario of the input
# files spanning the switch-over date from 0.2 to  0.1 degree resolution
# CMC GHRSST files.  Instead I will apply rule "if ANY input file post-dates
# switch-over date, use the 0.1 deg CMC files".
($options{ghrsst} eq 'by_date') and do {
  $options{ghrsst} = '0.2';
  foreach my $key (sort keys %dynanc_dates) {
    my ($YYYY, $MM, $DD, $ddd, $hh) = @{$dynanc_dates{$key}};
    (sprintf("%04d%02d%02d", $YYYY, $MM, $DD) >= '20170101') and do {
      $options{ghrsst} = '0.1';
      next;
    };
  }
};
($options{verbose} == 3) and pretty_dump('%options', %options);


# Now we have resolved which CMC GHRSST files to use, we can make PCFs
# ````````````````````````````````````````````````````````````````````
make_PCFs($tmpdir, \%sources, \%targets, %options);


# Build date/time tokenized templates for dynamic ancillary data
# ``````````````````````````````````````````````````````````````
# I want to use standard date/time tokens - to be consistent with other
# CSPP projects, to reuse existing perl functions, and to avoid ambiguity.
# And I am going to fill in here some tokens that I know are actually constant.
my $cmc_template = acspo_dynanc_tokenized($sources{sim}, 'reference sst block');
my $hdf_template = acspo_dynanc_tokenized($sources{sim}, 'atmospheric data');
$cmc_template =~ s/%H/12/g; $cmc_template =~ s/%M/00/g; $cmc_template =~ s/%S/00/g;
$hdf_template =~ s/%F/012/g;


# Define dynamic data remote & local locations and file naming conventions
# ````````````````````````````````````````````````````````````````````````
# The root of the local directory for dynamic ancillary data is either defined
# on the command line or is assigned a default value. I am appending the agreed
# subdirectory naming convention to this root and copying it across to a hash
# that contains much of what I need to know to fetch and store ancillary data.
# See http://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/ for directory structure.
# NOTE: The GDAS information is in here BUT it is not used for ACSPO.
my $local_repo = File::Spec -> catfile($options{dynanc}, '%Y_%m_%d_%j');
my %dynanc_loc = (GFSPROTO => 'http://',
                  GFSSITE => 'jpssdb.ssec.wisc.edu',
                  GFSREMOTE => 'cspp_v_2_0/ancillary/%Y_%m_%d_%j',
                  GFSTEMPLATE => 'gfs.t%H.%y%m%d.pgrbf12',
                  GFSLOCAL => $local_repo,
                  GDASPROTO => 'http://',
                  GDASSITE => 'jpssdb.ssec.wisc.edu',
                  GDASREMOTE => 'cspp_v_2_0/ancillary/%Y_%m_%d_%j',
                  GDASTEMPLATE => 'gdas1.pgrb00.1p0deg.%Y%m%d_%H_000.grib2',
                  GDASLOCAL => $local_repo,
                  CMCPROTO => 'http://',
                  CMCSITE => 'jpssdb.ssec.wisc.edu',
                  CMCREMOTE => 'cspp_v_2_0/ancillary/%Y_%m_%d_%j',
                  CMCTEMPLATE => $cmc_template,
                  CMCLOCAL => $local_repo,
                  HDFTEMPLATE => $hdf_template);
($options{verbose} == 3) and pretty_dump('%dynanc_loc', %dynanc_loc);


# Test for presence of dynamic ancillary data and go fetch if need be
# ```````````````````````````````````````````````````````````````````
my %dynanc_links = acspo_dynanc_management(\%dynanc_loc, %dynanc_dates);
($options{verbose} == 3) and pretty_dump('%dynanc_links', %dynanc_links);
foreach (sort keys %dynanc_links) {
  symlink($dynanc_links{$_}, $_) or
    die "$MAIN\:: Cannot make link to $dynanc_links{$_}\nAbort";
}


# Remainder of script only fully completed if --nothing not set on command line
# `````````````````````````````````````````````````````````````````````````````
# If the --noclean has also been set on command line, the temporary working
# directory will remain when this script terminates and in there will be a
# bash script  execACSPO.bash  that you could use for debugging, say.


# The executable for this MISSION is stored in $options{exe}
# ``````````````````````````````````````````````````````````
# I cannot get the following tee-ing to work with system(), so I simplified.
#   command > >(tee stdout.log) 2> >(tee stderr.log >&2) 
my ($stdout, $stderr) = ('> stdout.log', '2> stderr.log');
($options{verbose} >= 2) and $stdout = '';
($options{verbose} >= 3) and $stderr = '';
my $cmd = "(time $options{exe}) $stdout $stderr";
open(my $fh, '>execACSPO.bash') or die "$MAIN\:: Cannot open file execACSPO.bash $!";
print $fh "#!/bin/bash\n#ACSPO execution command\nulimit -s unlimited\n$options{exe}\n";
close($fh) or die "$MAIN\:: Cannot close file execACSPO.bash $!";
system('chmod 755 execACSPO.bash');
if ($options{nothing}) {
  ($options{verbose} >= 2) and print "--nothing so skipping $cmd execution\n";
} else {
  ($options{verbose} >= 2) and print "Executing $cmd\n";
  system($cmd) == 0 or die "$MAIN\:: Could not run command $cmd\nAbort";
}


# Copy the bias files from this run back to BIAS_DIR, and log files to WORK_DIR
# `````````````````````````````````````````````````````````````````````````````
if ($options{nothing}) {
  ($options{verbose} >= 2) and print "--nothing so skipping bias file copy\n";
  ($options{verbose} >= 2) and print "--nothing so skipping log file copy\n";
} else {
  ($options{verbose} >= 2) and print "Copying bias files to $options{bias}\n";
  foreach (sort(glob('*_HISTOGRAM_*.dat'))) {
    copy($_, $options{bias}) or do {
      ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!"; };
  }
  ($options{verbose} >= 2) and print "Copying log files to $options{work}\n";
  foreach (sort(glob('*.log'))) {
    copy($_, $options{work}) or do {
      ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!"; };
  }
}


# All done, chdir back to start dir (so $tmpdir can be removed) and exit
# ``````````````````````````````````````````````````````````````````````
if (not $options{clean}) {
  print "--noclean set, reminder: tmpdir is ...\n$tmpdir\n";
}
chdir($pwd) or die "$MAIN\:: Cannot chdir back to $pwd\nAbort";
exit(0);



###############################################################################
#                                                                             #
#         User-defined functions - they are listed alphabetically             #
#                                                                             #
#   NOTE: My editor can collapse code between matching {} braces, this        #
#   explains the strange formatting - I can collapse each function to see     #
#   just the function name and an example of how to call it; and expand       #
#   to see the more detailed documentation and to edit.      JED Nov 2013     #
#                                                                             #
###############################################################################

# ----------------------
  sub acspo_dynanc_dates
# ======================
# %dynanc_dates = acspo_dynanc_dates(%options);
  {
    # %options provides details of files to process in this run which, in turn,
    # tells me what specific dates of ancillary data I need.
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    # The files to be processed determine date/time for dynamic ancillary data
    my ($YYYY, $MM, $ddd, $DD, $hh, $mm, $ss, $YY) = ('N/A') x 8;
    my %dynanc_dates = (); # what is important is hash property of key uniqueness
    my $k = 0;
    foreach my $shortname (@{$options{process_files}}) {
      my $file = File::Spec -> catfile('input', $shortname);
      foreach my $qual (qw/Beginning Ending/) {
        my $key = "$shortname-$qual";
        ($options{sat} eq 'npp') and do {
          if ($options{xfiles}) {
            ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_gmtco($file, $qual);
          } else {
            ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_gmodo($file, $qual);
          }
        };
        (grep {$_ eq $options{sat}} qw/aqua terra/) and do {
          my $geo = File::Spec -> catfile('input', $options{mod03}[$k]);
          ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_mod03($geo, $qual);
        };
        (grep {$_ eq $options{sat}} qw/n15 n16 n17 n18 n19 metopA metopB/) and
          ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_l1b($file, $qual);
        is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm) or
          die "$name\: Invalid Date Time $YYYY-$MM-$DD $hh:$mm from $file\nAbort";
        ($YYYY, $MM, $DD, $ddd) = evaluate_plausible_date("$YYYY-$MM-$DD");
        $dynanc_dates{$key} = [($YYYY, $MM, $DD, $ddd, $hh)];
      }
      $k++; # This exists just to index @{$options{mod03}} 
    }
    return %dynanc_dates;
  }

# ---------------------------
  sub acspo_dynanc_management
# ===========================
# %dynanc_links = acspo_dynanc_management(\%dynanc_loc, %dynanc_dates);
  {
    # %dynanc_loc is a hash of ancillary meta-data about where and how to fetch
    # the ancillary data for this run. %dynanc_dates are specific dates of
    # ancillary data I need. The return value is a hash of links that need
    # to be made in the temporary working directory to the ancillary data. As
    # far as ACSPO is concerned, the temporary working directory IS the dynamic
    # ancillary data directory; as far as CSPP_ACSPO is concerned, the command
    # line provides the dynamic ancillary data directory, or a default value is
    # assigned in get_commandline(). $options{mod03} was added to hold MODIS
    # geo files needed for date/time.  See cspp_preprocess() for explanation.
    my %dynanc_loc = %{(shift)};
    my (%dynanc_dates) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    %dynanc_links = ();
    # The files to be processed determine date/time for dynamic ancillary data
    my ($YYYY, $MM, $ddd, $DD, $hh, $mm, $ss, $YY) = ('N/A') x 8;
    # %cmc_files has a date key; I need to check older dates because of staging latency
    my %cmc_files = ();
    foreach my $key (sort keys %dynanc_dates) {
      ($YYYY, $MM, $DD, $ddd, $hh) = @{$dynanc_dates{$key}};
      for (my $doy=$ddd; $doy>($ddd-7); $doy--) {
        my ($Y, $M, $D, $d, $h) = get_datetime_from_gmtime($YYYY, $doy, $hh);
        my $cmc_key = sprintf("%04d-%02d-%02d", $Y, $M, $D);
        my @datetime = ($Y, $M, $D, $d, $h);
        my ($url, $file, $local)  = pre_fetch_generic_acspo('CMC', @datetime, %dynanc_loc);
        $cmc_files{$cmc_key} = [($url, $file, $local)];
      }
    }    
    # %hdf_files, %gfs_files & %gdas_files have date/hour key 
    my (%hdf_files, %gfs_files, %gdas_files) = ((),(),());
    foreach my $key (sort keys %dynanc_dates) {
      ($YYYY, $MM, $DD, $ddd, $hh) = @{$dynanc_dates{$key}};
      my (@datetime, $url, $file, $local) = ((),);
      my $hour = $hh - $hh % 6; # round down to 0, 6, 12, 18
      my ($year, $doy) = ($YYYY, $ddd);
      foreach (0, 6) { # i.e. for the 6-hour boundary before start & that time +6 hours
        # use $_==0 with -Beginning in $key, $_==6 with -Ending in $key
        (($_ == 0) and index($key, '-Ending') > 1) and next;
        (($_ == 6) and index($key, '-Beginning') > 1) and next;
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #                                                             #
        #   ACSPO presently does not support GDAS in place of GFS     #
        #   I have retained the code in case this situation changes   #
        #                                                             #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        ### # Correct for GDAS, where met. data is to bound satellite data time.
        ### ($YYYY, $MM, $DD, $ddd, $hh) = get_datetime_from_gmtime($year, $doy, $hour+$_);
        ### # This is the same one $key for %hdf_files, %gfs_files, %gdas_files
        ### my $key = sprintf("%04d-%02d-%02d-%02d", $YYYY, $MM, $DD, $hh);
        ### @datetime = ($YYYY, $MM, $DD, $ddd, $hh);
        ### ($url, $file, $local)  = pre_fetch_generic_acspo('GDAS', @datetime, %dynanc_loc);
        ### $gdas_files{$key} = [($url, $file, $local)];
        # GFS is the 12 hour forecast so the date/time is 12 hours earlier.
        ($YYYY, $MM, $DD, $ddd, $hh) = get_datetime_from_gmtime($year, $doy, $hour+$_-12);
        $YY = substr($YYYY, 2, 2);
        @datetime = ($YYYY, $MM, $DD, $ddd, $hh);
        ($url, $file, $local)  = pre_fetch_generic_acspo('GFS', @datetime, %dynanc_loc);
        $gfs_files{$key} = [($url, $file, $local)];
        # HDF is tagged for the 12 hour forecast, so date/time is same as GFS.
        $hdf_files{$key} = $dynanc_loc{HDFTEMPLATE};
        $YY = substr($YYYY, 2, 2);
        my @tokens = qw/%Y %y %m %d %j %H/;
        my @values = ($YYYY, $YY, $MM, $DD, $ddd, $hh);
        for (my $j=0; $j<=$#tokens; $j++) {
          $hdf_files{$key} =~ s/$tokens[$j]/$values[$j]/g;
        }
      }
    }
    ($options{verbose} == 3) and pretty_dump('%cmc_files', %cmc_files);
    ($options{verbose} == 3) and pretty_dump('%gfs_files', %gfs_files);
    ($options{verbose} == 3) and pretty_dump('%hdf_files', %hdf_files);
    #($options{verbose} == 3) and pretty_dump('%gdas_files', %gdas_files);
    # Get the CMC SST files
    ($options{verbose} >= 2) and print "$name\: INFO Fetching CMC files\n";
    foreach my $key (reverse(sort keys %cmc_files)) {
      ($options{verbose} == 3) and 
        print "$name\: DEBUG Fetching @{$cmc_files{$key}}[1]\n";
        $options{cmc_date} = $key;
        my $status = lftp_mirror(@{$cmc_files{$key}}, %options);
        ($status) and last; # got one, let's split
    }
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                                                                 #
    #  This is presently not supported by $GRIB2HDF external utility  #
    #  The developers at NOAA advise that GDAS not supported anyway   #
    #                                                                 #
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #
    # # Get the GDAS files and convert to HDF format
    # ($options{verbose} >= 2) and print "$name\: INFO Fetching GDAS files\n";
    # foreach my $key (sort keys %hdf_files) {
    #   my ($url, $file, $local) = @{$gdas_files{$key}};
    #   my $gdas = File::Spec -> catfile($local, $file);
    #   my $hdf = File::Spec -> catfile($local, $hdf_files{$key});
    #   -e $hdf and next;
    #   ($options{verbose} == 3) and print "$name\: DEBUG Fetching $file\n";
    #   my $status = lftp_mirror(@{$gdas_files{$key}}, %options);
    #   ($options{verbose} == 3) and print "$name\: DEBUG grib2hdf $file\n";
    #   my $cmd = "$GRIB2HDF 3 $gdas >> gdas_converter.log";
    #   system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
    #   $cmd = "mv $gdas.hdf $hdf";
    #   system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
    # }
    # Get the GFS files and convert to HDF format
    ($options{verbose} >= 2) and print "$name\: INFO Fetching GFS files\n";
    foreach my $key (sort keys %hdf_files) {
      my ($url, $file, $local) = @{$gfs_files{$key}};
      my $gfs = File::Spec -> catfile($local, $file);
      my $hdf = File::Spec -> catfile($local, $hdf_files{$key});
      -e $hdf and next;
      ($options{verbose} == 3) and print "$name\: DEBUG Fetching $file\n";
      my $status = lftp_mirror(@{$gfs_files{$key}}, %options);
      -e $gfs or die "$name\: $gfs not downloaded by $LFTP\nAbort";
      my $cmd = "$GRIB2HDF 3 $gfs >> gfs_converter.log";
      ($options{verbose} == 3) and print "$name\: DEBUG grib2hdf $file\n";
      system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
      -e $hdf and next; # put there by another process while you were busy?
      $cmd = "mv $gfs.hdf $hdf";
      system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
    }
    foreach (sort keys %cmc_files) {
      my ($url, $file, $local) = @{$cmc_files{$_}};
      my $check = File::Spec -> catfile($local, $file);
      -e $check and $dynanc_links{$file} = $check;
    }
    foreach (sort keys %hdf_files) { # same key for %hdf_files and %gfs_files
      my ($url, $file, $local) = @{$gfs_files{$_}};
      $file = $hdf_files{$_};
      $dynanc_links{$file} = File::Spec -> catfile($local, $file);
    }
    return %dynanc_links;
  }

# --------------------------
  sub acspo_dynanc_tokenized
# ==========================
# $tokenized = acspo_dynanc_tokenized($configfile, $blockname);
  {
    # Returns $tokenized, a tokenized template (with date/time tokens) for
    # managing dymanic ancillary data files. fname_pattern and time_template
    # are grepped out of block $blockname in $configfile.
    # The ACSPO system, as delivered, uses fname_pattern and time_template as
    # keywords in simDataConfigFile.txt (the value I have in $sources{sim}).
    # To be compliant with the date/time token expansion system I am using
    # I need to combine fname_pattern and time_template into a single,
    # tokenized return value, $tokenized.  The tokens I use are listed in the
    # header of function token_value(), near end of this script. The tokens
    # used in $time_template are yyyy yy MM dd hh mm ss fff.  NOTE you need
    # to test for longer acpso tokens before shorter (e.g. yyyy before yy)
    # otherwise the token replacement will mess up! (This is why CSPP uses
    # tokens that are NOT ambiguous, e.g. %Y for yyyy and %y for yy).
    my ($configfile, $blockname) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($fname_pattern, $time_template) = ('N/A', 'N/A');
    open(my $fh, "<$configfile");
    while (<$fh>) { chomp;
      if ( /start $blockname/ ) {
        (/#/) and next;
        while (<$fh>) { chomp;
          (/#/) and next;
          (/fname_pattern/) and $fname_pattern = (split(/\=/, $_))[1];
          (/time_template/) and $time_template = (split(/\=/, $_))[1];
          ( /end $blockname/ ) and last;
        }
      }
    }
    close $fh;
    ($fname_pattern eq 'N/A') and
      die "$name\: Could not find fname_pattern in $blockname\nAbort";
    ($time_template eq 'N/A') and
      die "$name\: Could not find time_template in $blockname\nAbort";
    my @replac = qw/yyyy yy MM ddd dd hh ss mm fff ff/; # ACSPO tokens
    my @tokens = qw/  %Y %y %m  %j %d %H %S %M  %F %f/;
    my @widths =   (   4, 2, 2,  3, 2, 2, 2, 2,  3, 2);
    my ($tokenized, $template) = ($fname_pattern, $time_template);
    for (my $j=0; $j<=$#replac; $j++) {
      my @parts = split(/$replac[$j]/, $template);
      ($#parts == 0) and next;
      my $tmp1 = $parts[0];
      my $tmp2 = substr($tokenized, 0, length($tmp1));
      for (my $i=1; $i<=$#parts; $i++) {
        my $m = length($tmp1) + $widths[$j];
        $tmp1 = $tmp1.$tokens[$j].$parts[$i];
        my $l = length($parts[$i]);
        $tmp2 = $tmp2.$tokens[$j].substr($tokenized, $m, $l);
      }
      $template = $tmp1;
      $tokenized = $tmp2;
    }
    return $tokenized;
  }

# -------------------
  sub cspp_preprocess
# ===================
# %options = cspp_preprocess(%options);
  {
    # On input, %options comprises some command line parameters plus other
    # processing options, including defaults for unspecified command line values
    # and extra stuff defined in main to guide processing. On output, %options
    # needs to include a field $options{process_files} that points to files
    # that satisfy ACSPO system requirements in terms of both filename and
    # internal format.
    #
    # In this function I will make the input directory and populate it with the
    # input data, or links to the input data, as appropriate.  The reason I do
    # not simply link to the user supplied input directory is that the input
    # data may not be in the format I require, either internally or in terms of
    # file naming convention.
    #
    # For npp VIIRS data, I use the aggregator in the NOAA delivery to create
    # concatenated, swath length files (if --aggregate, which is the default).
    #
    # The files for 1bFilesList.txt I put in $options{process_files}. npp VIIRS
    # uses geolocation files as the files for 1bFilesList.txt but, annoyingly,
    # aqua/terra MODIS uses the 1KM radiance files. I had been interrogating 
    # (internally) @{$options{process_files}} for date/time information, now
    # I need a separate list of the MODIS geo files for that purpose. I am
    # going to have these put in $options{mod03}.
    #
    # For npp and aqua/terra there is a pre-processing step of destriping in
    # ACSPO v2.40 which now needs to be slotted in. This occurs in-place, so
    # better not have links to data files, but copies.
    #
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @process_files = ();
    my @mod03 = ();
    mkpath('input') or die "$name\: Cannot make directory 'input'\nAbort";
    # npp VIIRS processing
    ($options{sat} eq 'npp') and do {
      # copy destriping parameter file to temporary working directory
      my $source = File::Spec -> catfile($HOME, 'run', 'static_anc',
      'viirs', 'destriping_viirs_param.txt');
      copy($source, '.') or die "$name\: Cannot copy $source.\nAbort";
      # a links directory simplifies things because we can later refer to the
      # whole directory without worrying about including files we do not want
      mkpath('links') or die "$name\: WORK_DIR/links cannot be made.\nAbort";
      # make symlinks to just the files we want to process
      foreach (@{$options{input_files}}) {
        my $target = File::Spec -> catfile($options{input_dir}, $_);
        my $linkname = File::Spec -> catfile('links', $_);
        symlink($target, $linkname) or die "$name\: Cannot make link to $target\nAbort";
      }
      # aggregate or not, we must first disaggregate with nagg
      mkpath('disagg') or die "$name\: Cannot make directory 'disagg'\nAbort";
      my $rad = 'SVM05,SVM07,SVM10,SVM12,SVM15,SVM16';
      my $geo_type = 'GMODO';
      ($options{xfiles}) and $geo_type = 'GMTCO';
      my @fields = split(/\_/, basename(@{$options{input_files}}));
      my $origin = $fields[7];
      my $domain = (split(/\./, basename($fields[8])))[0];
      my $cmd = "$H5NAGG -S -g $geo_type -t $rad -n 1 -d disagg -O $origin -D $domain links/*";
      my ($stdout, $stderr) = ('> stdout.nagg', '2> stderr.nagg');
      ($options{verbose} >= 3) and $stdout = '';
      $cmd = "$cmd $stdout $stderr";
      system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
      my $globber = File::Spec -> catfile('disagg', '*.h5');
      my @input_files = ();
      foreach (sort(glob($globber))) { push(@input_files, basename($_)); }
      if ($options{aggregate}) {
        my %swaths = npp_granules_to_swaths(@input_files);
        ($options{verbose} == 3) and pretty_dump('%swaths', %swaths);
        my $IET = "$HOME/run/static_anc/viirs/IETTime.dat";
        copy($IET, '.') or die "$name\: Copy $IET failed: $!\nAbort";
        my $iswath = 0;
        foreach (sort keys %swaths) {
          my $pcf = sprintf("aggCntlFile_%03d.txt", ++$iswath);
          make_aggCntlFile($pcf, \%options, @{$swaths{$_}});
          my $log = sprintf("aggCntlFile_%03d.log", $iswath);
          my $cmd = "$VIIRSAGG <$pcf >$log";
          system($cmd) == 0 or warn "$name\: Could not run command $cmd\n";
        }
      } else {
        foreach (@input_files) {
          my $target = File::Spec -> catfile('..','disagg', $_);
          my $linkname = File::Spec -> catfile('input', $_);
          symlink($target, $linkname) or die "$name\: Cannot make link to $target\nAbort";
        }
      }
      my @geo = sort(glob('input/GMODO*'));
      ($options{xfiles}) and @geo = sort(glob('input/GMTCO*'));
      ($#geo > -1) or die "$name\: Nothing found to process\nAbort";
      my @svm = sort(glob('input/SVM*'));
      my $n = ($#svm+1)/($#geo+1);
      is_integer($n) or die "$name\: Mis-match in SVM/GEO file count\nAbort";
      ($n > 5) or die "$name\: Too few SVM bands for algorithm\nAbort";
      foreach (grep /SVM1[256]/, @svm) {
        my $cmd = "$options{destripe} $_ destriping_viirs_param.txt";
        system($cmd.' >> vdestripe.log 2>&1') == 0 or
          die "$name\: Could not run command $cmd\nAbort";
      }
      foreach (@geo) { push(@process_files, basename($_)); }
    };
    # aqua/terra MODIS processing
    (grep {$options{sat} eq $_} qw/aqua terra/) and do {
      # copy destriping parameter file to temporary working directory
      my $source = File::Spec -> catfile($HOME, 'run', 'static_anc',
      'modis', 'destriping_modis_param.txt');
      copy($source, '.') or die "$name\: Cannot copy $source.\nAbort";
      foreach (qw|/run/modis/filetable.temp_cspp /run/static_anc/modis/leapsec.dat_cspp|) {
        my $target = basename($_);
        $target =~ s/_cspp//g;
        copy($HOME.$_, $target) or die "$name\: Copy $HOME$_ failed: $!\nAbort";
      }
      foreach (@{$options{input_files}}) {
        my $target = File::Spec -> catfile($options{input_dir}, $_);
        my $newname = File::Spec -> catfile('input', $_);
        ($options{xfiles}) or do {
          my $shortname = get_filename_from_CoreMetadata($target);
          $newname = File::Spec -> catfile('input', $shortname); };
        copy($target, $newname) or
          die "$name\: Cannot copy $newname to $target\nAbort";
      }
      my @rad = sort(glob('input/M?D021KM.*'));
      my @geo = sort(glob('input/M?D03.*'));
      ($#geo < 0) and die "$name\: $options{sat} GEO files not found.\nAbort";
      ($#rad < 0) and die "$name\: $options{sat} RAD files not found.\nAbort";
      foreach (0..$#geo) {
        my $rad_collect = substr($rad[$_],29,3); # count includes input/
        my $geo_collect = substr($geo[$_],26,3);
        ($geo_collect eq $rad_collect) or do {
          my $target = $geo[$_];
          substr($geo[$_],26,3) = $rad_collect;
          symlink(basename($target), $geo[$_]) or
            die "$name\: Cannot make $geo[$_] link to $target\nAbort";
        };
        my $cmd = "$options{destripe} $rad[$_] destriping_modis_param.txt";
        system($cmd.' >> mdestripe.log 2>&1') == 0 or
          die "$name\: Could not run command $cmd\nAbort";
      }
      foreach (@rad) { push(@process_files, basename($_)); }
      foreach (@geo) { push(@mod03, basename($_)); }
      symlink('input', 'destripe') or
        die "$name\: Cannot make input/ link to destripe/\nAbort";
    };
    # n15/n16/n17/n18/n19/metopA/metopB AVHRR processing
    (grep {$options{sat} eq $_} qw/n15 n16 n17 n18 n19 metopA metopB/) and do {
      foreach (@{$options{input_files}}) {
        my $target = File::Spec -> catfile($options{input_dir}, $_);
        my $new_target = make_avhrr_L1B_bigEndian($target, 'input');
        $options{site_id} = substr($new_target, 0, 3);
        my $linkname = 'NSS'.substr($new_target, 3);
        $linkname =~ s/\.MB\./\.M1\./; $linkname =~ s/\.MA\./\.M2\./;
        ($linkname eq $new_target) or do { # avoid link-to-self (which will fail!)
          my $full_link = File::Spec -> catfile('input', $linkname);
          symlink($new_target, $full_link) or
            die "$name\: Cannot link $linkname to $new_target\nAbort"; };
        push(@process_files, $linkname);
      } 
    };
    @{$options{mod03}} = @mod03; # empty list except for aqua/terra MODIS files
    @{$options{process_files}} = @process_files;
    return %options;
  }

# ---------------------------
  sub evaluate_plausible_date
# ===========================
# ($YYYY, $MM, $DD, $ddd) = evaluate_plausible_date($datestring);
  {
    # Input string IS one of these with sensible integer component values:
    #  %Y%m%d %Y%j %Y-%m-%d %Y-%j YESTERDAY TODAY, unpack & return %Y %m %d %j
    # NOTE: Assumes is_plausible_datetime() has already affirmed!
    # This is a both ugly and more general than necessary - because it was
    # filched from an IMAPP project, namely IDEA-I.
    my @parts = split(/-/, $_[0]);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($YYYY, $MM, $DD, $ddd) = undef;
    my $last_yday = undef;
    if ($#parts == 0) {
      my $form = $parts[0];
      if (is_integer($form)) {
        if (length($form) == 7) {     # $form = YYYYddd
          $YYYY = substr($form,0,4);
          $ddd = substr($form,4,3);
        }
        if (length($form) == 8) {     #  $form = YYYYMMDD
          $YYYY = substr($form,0,4);
          $MM = substr($form,4,2);
          $DD = substr($form,6,2);
        }
      } else {
        my $dayword = uc($form);
        (grep {$_ eq $dayword} qw/YESTERDAY TODAY TOMORROW/) and do {
          gmtime();
          $YYYY = 1900 + $tm_year;
          $ddd = $tm_yday + 1; # TODAY
          ($dayword eq 'YESTERDAY') and do {
            (--$ddd == 0) and $ddd = days_in_year(--$YYYY); };
          ($dayword eq 'TOMORROW') and do {
            ($ddd++ == days_in_year($YYYY)) and do {
              $ddd = 1; $YYYY++; };
          };
        };
      }
    }
    if ($#parts == 1) {
      ($YYYY, $ddd) = @parts;
    }
    if ($#parts == 2) {
      ($YYYY, $MM, $DD) = @parts;
    }
    my @days = (31,28,31,30,31,30,31,31,30,31,30,31);
    is_leap_year($YYYY) and $days[1] = 29;
    if (defined($ddd)) {         # convert ddd -->  MM, DD
      ($MM, $DD) = (0, 0);
      while ((sum(@days[0..$MM]) < $ddd) and ($MM < 11)) {
        $DD += $days[$MM];
        $MM++;
      }
      $MM++;
      $DD = $ddd - $DD;
    } else {                     # convert MM, DD -->  ddd
      if ($MM > 1) {
        $ddd = sum(@days[0..($MM-2)])
      }
      $ddd += $DD;
    }
    if ($ddd > sum(@days)) {
      print "Year:$YYYY Month:$MM Day:$DD DOY:$ddd\n";
      die "$name\: Day of year > number of days in year\nAbort";
    }
    if ($DD > $days[$MM-1]) {
      print "Year:$YYYY Month:$MM Day:$DD DOY:$ddd\n";
      die "$name\: Day of month > number of days in month\nAbort";
    }
    $YYYY = sprintf("%04d", $YYYY);
    $MM = sprintf("%02d", $MM);
    $DD = sprintf("%02d", $DD);
    $ddd = sprintf("%03d", $ddd);
    return ($YYYY, $MM, $DD, $ddd);
  }

# ---------------------------
  sub first_doublequoted_word
# ===========================
# $word = first_doublequoted_word($string);
  {
    # $string contains a double quoted word, for example,
    # "MYD03.A2015028.1834.006.2015036084827.hdf"\012,
    # that you want to pull out from between the double quotes.
    # This is a utility routine for get_filename_from_CoreMetadata().
    # Returns the first word, unquoted, or 'N/A'.
    my $string = shift;
    my $fc = index($string, '"');
    ($fc == -1) and return 'N/A';
    $fc++;
    my $lc = index($string, '"', $fc);
    ($lc == -1) and return 'N/A';
    my $nc = $lc - $fc;
    my $word = substr($string, $fc, $nc);
    return $word;
  }

# -------------------
  sub get_commandline
# ===================
# %options = get_commandline($revision, $pwd);
  {
    # If things go bad in here, jump to usage() which terminates with exit(9).
    # Note that I had to hack in the Python-like -vv & -vvv support.
    # This routine assumes standard module Getopt::Long has been loaded.
    # abs_path() is dangerous for the purpose of setting the work (and,
    # equivalently, the output) directory - see this post:
    # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=257568
    my ($rev, $pwd) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $codeset = langinfo(CODESET);
    @ARGV = map{decode $codeset, $_} @ARGV;
    foreach (@ARGV) { s/\?/-\?/g; }
    my %options = ();
    GetOptions(\%options,
      'version|V',
      'help|?',
      'nothing|n',
      'input_files|i=s',
      'sat|s=s',
      'bias|b=s',
      'dynanc|d=s',
      'work|w=s',
      'ghrsst|r=s',
      'processors|p=i',
      'verbose|v+', 'vv', 'vvv',
      'xfiles|x!',
      'granule|g!',
      'aggregate|a',
      'quarantine|q!',
      'clean|c!') or usage();
    defined($options{version}) and die "$rev\n";
    defined($options{help}) and usage();
    $options{nothing} = defined($options{nothing}) ? $options{nothing} : 0;
    $options{xfiles} = defined($options{xfiles}) ? $options{xfiles} : 0;
    $options{ghrsst} = defined($options{ghrsst}) ? $options{ghrsst} : 'by_date';
    defined($options{input_files}) or usage();
    defined($options{sat}) or usage();
    defined($options{bias}) or usage();
    (grep {$_ eq $options{sat}} qw/aqua terra npp n15 n16 n17 n18 n19 metopA metopB/) or usage();
    (defined($options{work})) or $options{work} = $pwd;
    (defined($options{dynanc})) or
      $options{dynanc} = File::Spec -> catfile($ENV{CSPP_ACSPO_HOME}, 'data', 'dynamic_ancillary');
    foreach (qw/bias dynanc work/) {
      (substr($options{$_}, 0, 1) eq '/') or do {
        $options{$_} = File::Spec -> catfile($pwd, $options{$_}); };
    }
    $options{processors} = defined($options{processors})
      ? $options{processors} : 1; # not currently used
    $options{verbose} = defined($options{verbose}) ? $options{verbose} : 0;
    $options{vv} = defined($options{vv}) ? $options{vv} : 0;
    $options{vvv} = defined($options{vvv}) ? $options{vvv} : 0;
    $options{verbose} = $options{verbose} + 2 * $options{vv};
    $options{verbose} = $options{verbose} + 3 * $options{vvv};
    $options{verbose} = min($options{verbose}, 3);
    delete @options{'vv', 'vvv'};
    $options{clean} = defined($options{clean}) ? $options{clean} : 1;
    $options{granule} = defined($options{granule}) ? $options{granule} : 0;
    $options{quarantine} = defined($options{quarantine}) ? $options{quarantine} : 0;
    $options{aggregate} = defined($options{aggregate}) ? $options{aggregate} : 1;
    return %options;
  }

# ---------------------------
  sub get_datetime_from_gmodo
# ===========================
# ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_gmodo($file, $block);
  {
    # $file is the name of the GMODO file, full path if not in tempdir.
    # $block is either 'Beginning' or 'Ending' - ok, you get the picture.
    # Uses global $H5DUMP as validated location of h5dump executable
    # This is a filename independent way of getting date/time from
    # GMODO h5 file. Is it better than using the filename? - maybe not
    # since the filenaming convention for npp SDRs is well framed.
    #
    my ($file, $block) = (shift, shift);
    my $var = 'Data_Products/VIIRS-MOD-GEO/VIIRS-MOD-GEO_Gran_0';
    my $block_Date = '"'.$block.'_Date"';
    my $block_Time = '"'.$block.'_Time"';
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($date, $time) = ('N/A', 'N/A');
    my $cmd = "$H5DUMP --onlyattr -d $var $file |";
    open(my $fh, $cmd) or die "$name\: Could not run command $cmd\nAbort";
    while (<$fh>) { chomp;
      /$block_Date/ and do {
        while (<$fh>) { chomp;
          /0\)\:/ and do {
            my $tmp = (split(/0\)\:/, $_))[1];
            $tmp =~ s/\"//g;
            $date = trim_string($tmp); 
            last; };
        } };
      /$block_Time/ and do {
        while (<$fh>) { chomp;
          /0\)\:/ and do {
            my $tmp = (split(/0\)\:/, $_))[1];
            $tmp =~ s/\"//g;
            $time = trim_string($tmp); 
            last; };
        } };
    }
    close($fh);
    my $YYYY = substr($date, 0, 4);
    my $MM = substr($date, 4, 2);
    my $DD = substr($date, 6, 2);
    my $hh = substr($time, 0, 2);
    my $mm = substr($time, 2, 2);
    my $ss = substr($time, 4, 2);
    return ($YYYY, $MM, $DD, $hh, $mm, $ss);
  }

# ---------------------------
  sub get_datetime_from_gmtco
# ===========================
# ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_gmtco($file, $block);
  {
    # $file is the name of the GMTCO file, full path if not in tempdir.
    # $block is either 'Beginning' or 'Ending' - ok, you get the picture.
    # Uses global $H5DUMP as validated location of h5dump executable
    # This is a filename independent way of getting date/time from
    # GMTCO h5 file. Is it better than using the filename? - maybe not
    # since the filenaming convention for npp SDRs is well framed.
    #
    my ($file, $block) = (shift, shift);
    my $var = 'Data_Products/VIIRS-MOD-GEO-TC/VIIRS-MOD-GEO-TC_Gran_0';
    my $block_Date = '"'.$block.'_Date"';
    my $block_Time = '"'.$block.'_Time"';
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($date, $time) = ('N/A', 'N/A');
    my $cmd = "$H5DUMP --onlyattr -d $var $file |";
    open(my $fh, $cmd) or die "$name\: Could not run command $cmd\nAbort";
    while (<$fh>) { chomp;
      /$block_Date/ and do {
        while (<$fh>) { chomp;
          /0\)\:/ and do {
            my $tmp = (split(/0\)\:/, $_))[1];
            $tmp =~ s/\"//g;
            $date = trim_string($tmp); 
            last; };
        } };
      /$block_Time/ and do {
        while (<$fh>) { chomp;
          /0\)\:/ and do {
            my $tmp = (split(/0\)\:/, $_))[1];
            $tmp =~ s/\"//g;
            $time = trim_string($tmp); 
            last; };
        } };
    }
    close($fh);
    my $YYYY = substr($date, 0, 4);
    my $MM = substr($date, 4, 2);
    my $DD = substr($date, 6, 2);
    my $hh = substr($time, 0, 2);
    my $mm = substr($time, 2, 2);
    my $ss = substr($time, 4, 2);
    return ($YYYY, $MM, $DD, $hh, $mm, $ss);
  }

# ----------------------------
  sub get_datetime_from_gmtime
# ============================
# ($YYYY, $MM, $DD, $ddd, $hh) = get_datetime_from_gmtime($year, $doy, $hour);
  {
    # A utility for use with date/time arithmetic, limited to whole hours.
    # Because timegm_nocheck() is used, you can put in $doy (day of year) or
    # $hour that are outside normal bounds (like negative) and a sensible date
    # and time are returned (so a -ve $doy will end up giving you a +ve $ddd in
    # a previous year $YYYY). This is a way to get the system to do the logic
    # when it comes to date/time arithmetic. I have limited precison to whole
    # hours, because that is all I need in this project.
    my ($year, $doy, $hour) = (shift, shift, shift);
    my $gmtime = timegm_nocheck(0, 0, $hour, $doy, 0, $year);
    gmtime($gmtime);
    my $YYYY = sprintf("%04d", $tm_year + 1900);
    my $MM = sprintf("%02d", $tm_mon + 1);
    my $DD = sprintf("%02d", $tm_mday);
    my $ddd = sprintf("%03d", $tm_yday + 1);
    my $hh = sprintf("%02d", $tm_hour);
    return ($YYYY, $MM, $DD, $ddd, $hh);
  }

# -------------------------
  sub get_datetime_from_l1b
# =========================
# ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_l1b($file, $block);
  {
    # $file is the name of the l1b file, full path or relative to temporary
    # working directory (which is $PWD). $file is bigEndian.
    # $block is either 'Beginning' or 'Ending' - ok, you get the picture.
    # This is a filename independent way of getting date/time from
    # AAPP/NOAA L1B file. Is it better than using the filename? - maybe,
    # because I am supporting different filename conventions.  But, by the
    # time this routine is called, the l1b file is going to be named like:
    # $PWD/input/NSS.HRPT.NN.D13276.S0719.E0724.B4313131.WI
    #
    my ($file, $block) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($count, $buf) = undef;
    my ($year, $doy, $msec) = undef;
    my $reclen = 112; # only read start of header; this keeps it generic
    my $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv6';
    my $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g;
    open my $fh, "<$file" or die "$name\: open $file for read failed\nAbort";
    binmode $fh;
    $count = read($fh, $buf, $reclen) or die "$name\: read $file failed\nAbort";
    ($count == $reclen) or die "$name\: Header count != reclen $file\nAbort";
    my @data = unpack($Bhdr, $buf);
    close $fh;
    ($block eq 'Beginning') and ($year, $doy, $msec) = ($data[18], $data[19], $data[20]);
    ($block eq 'Ending')    and ($year, $doy, $msec) = ($data[22], $data[23], $data[24]);
    my $gmtime = timegm_nocheck($msec/1000, 0, 0, $doy, 0, $year);
    gmtime($gmtime);
    my $YYYY = sprintf("%04d", $tm_year + 1900);
    my $MM   = sprintf("%02d", $tm_mon + 1);
    my $DD   = sprintf("%02d", $tm_mday);
    my $ddd  = sprintf("%03d", $tm_yday + 1);
    my $hh   = sprintf("%02d", $tm_hour);
    my $mm   = sprintf("%02d", $tm_min);
    my $ss   = sprintf("%02d", $tm_sec);
    return ($YYYY, $MM, $DD, $hh, $mm, $ss);
  }

# ---------------------------
  sub get_datetime_from_mod03
# ===========================
# ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_mod03($file, $block);
  {
    # $file is the name of the MOD03 file, full path if not in tempdir.
    # $block is either 'Beginning' or 'Ending' - ok, you get the picture.
    # Uses global $HDP as validated location of h5dump executable
    # This is a filename independent way of getting date/time from
    # MOD03 hdf file. Is it better than using the filename? - maybe,
    # because I am supporting different filename conventions AND the
    # conventions (unlike S-NPP file) do not give end-of-data time.
    #
    my ($file, $block) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($date, $time) = ('N/A', 'N/A');
    my ($YYYY, $MM, $DD, $hh, $mm, $ss) = ('N/A') x 6;
    my $var =  "\'SD start time\'";
    my @seconds = split(/\s+/, `$HDP dumpsds -s -d -n $var $file`);
    ($block eq 'Ending') and @seconds = reverse(@seconds);
    foreach my $t (@seconds) {
      my $gmtime = timegm_nocheck($t, 0, 0, 1, 0, 1993);
      gmtime($gmtime);
      $YYYY = sprintf("%04d", $tm_year + 1900);
      $MM = sprintf("%02d", $tm_mon + 1);
      $DD = sprintf("%02d", $tm_mday);
      $hh = sprintf("%02d", $tm_hour);
      $mm = sprintf("%02d", $tm_min);
      $ss = sprintf("%02d", $tm_sec);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm) and last;
    }
    return ($YYYY, $MM, $DD, $hh, $mm, $ss);
  }

# ----------------------------------
  sub get_filename_from_CoreMetadata
# ==================================
# $filename = get_filename_from_CoreMetadata($file);
  {
    # $file is the path/filename of aqua/terra hdf file (seen from tempdir).
    # Gets the filename in M?D{03,021KM}.YYYYddd.hhmm.etc form by interrogating
    # CoreMetadata in aqua/terra hdf file. This is a bit painful because the
    # complete geo filename is in the geo file, but the rad filename is not in
    # the 1km radiance file - so we have to hack a solution to the latter.
    my $file = shift;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $cmd = "$HDP dumpsds -s -h $file |";
    open(my $fh, $cmd) or die "$name\: Could not run command $cmd\nAbort";
    my $filename = 'filename_undiscovered_by_get_filename_from_CoreMetadata';
    my ($line, $found) = ('N/A', 0);
    my @words = ();
    while (<$fh>) { chomp;
      ( /Attr/ and /CoreMetadata/ ) and do {
        do { $line = <$fh>; $found = (index($line, 'Value') > -1); } until ($found or eof);
        chomp $line; ($found) and @words = split(/\s+/, $line); };
    }
    close($fh);
    my ($date, $time, $prod, $type, $value, $collect) = ('N/A') x 6;
    my $k = 0;
    foreach my $word (@words) {
      foreach (qw/V5. V6./) {
        (index($word, $_) > -1 and $collect eq 'N/A') and do {
          /5/ and $collect = '005';
          /6/ and $collect = '006'; };
      }
      foreach (qw/MOD03 MYD03 MOD021KM MYD021KM/) {
        (index($word, $_) > -1 and $type eq 'N/A') and
          ($type, $value) = ($_, first_doublequoted_word($word));
      }
      (index($word, 'PRODUCTIONDATETIME') > -1 and $prod eq 'N/A') and
        $prod = first_doublequoted_word($words[$k+6]);
      (index($word, 'RANGEBEGINNINGDATE') > -1 and $date eq 'N/A') and
        $date = first_doublequoted_word($words[$k+6]);
      (index($word, 'RANGEBEGINNINGTIME') > -1 and $time eq 'N/A') and
        $time = first_doublequoted_word($words[$k+6]);
      $k++
    }
    (grep {$_ eq $type} qw/MOD03 MYD03/) and return $value;
    ($date eq 'N/A') and return $filename;
    ($time eq 'N/A') and return $filename;
    ($type eq 'N/A') and return $filename;
    ($prod eq 'N/A') and $prod='2037-12-31T23:59:59.999Z'; # not worth failing over
    ($collect eq 'N/A') and return $filename;
    (grep {$_ eq $type} qw/MOD021KM MYD021KM/) and do {
      my $YYYY = substr($date,  0, 4);
      my $MM   = substr($date,  5, 2);
      my $DD   = substr($date,  8, 2);
      my $hh   = substr($time,  0, 2);
      my $mm   = substr($time,  3, 2);
      my $ss   = substr($time,  6, 2);
      my $gmtime = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      gmtime($gmtime);
      my $ddd = sprintf("%03d", $tm_yday + 1);
      my $head = "$type.A$YYYY$ddd.$hh$mm.$collect";
      $YYYY = substr($prod,  0, 4);
      $MM   = substr($prod,  5, 2);
      $DD   = substr($prod,  8, 2);
      $hh   = substr($prod, 11, 2);
      $mm   = substr($prod, 14, 2);
      $ss   = substr($prod, 17, 2);
      $gmtime = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      gmtime($gmtime);
      $ddd = sprintf("%03d", $tm_yday + 1);
      my $tail = "$YYYY$ddd$hh$mm$ss.hdf";
      $filename = "$head.$tail";
    };
    return $filename;
  }

# -----------------------------------------------------------
  sub is_integer { defined $_[0] and $_[0] =~ /^[+-]?\d+$/; }
# ===========================================================

# ----------------
  sub is_leap_year
# ================
# $status = is_leap_year($YYYY);
  {
    # Accepts four-digit year, return 1 if leap year, 0 otherwise.
    my $YYYY = $_[0];
    my $leap = 0;
    ($YYYY % 4) == 0 and $leap = 1;
    ($YYYY % 100) == 0 and ($YYYY % 400) != 0 and $leap = 0;
    return $leap;
  }

# -------------------------
  sub is_plausible_datetime
# =========================
# $status = is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm);
  {
    # NOTE: Year < 2000 is considered implausible in this function.
    my ($YYYY, $MM, $DD, $hh, $mm) = (@_);
    my $implausible_year = 2000;
    is_integer($YYYY) or return 0;
    $YYYY < $implausible_year and return 0;
    is_integer($MM) or return 0;
    $MM >= 1 and $MM <= 12 or return 0;
    is_integer($DD) or return 0;
    $DD >= 1 and $DD <= 31 or return 0;
    is_integer($hh) or return 0;
    $hh >= 0 and $hh <= 23 or return 0;
    is_integer($mm) or return 0;
    $mm >= 0 and $mm <= 59 or return 0;
    return 1;
  }

# ---------------
  sub lftp_mirror
# ===============
# $status = lftp_mirror($site, $file, $local, %options);
  {
    # Executes lftp in mirror mode, but if NOT $options{quarantine}.
    # To avoid clashes where multiple fetches of the same data to the
    # same ancillary data location occur simultaneously, I now fetch
    # data to temporary working directory an copy back to permanent
    # dynamic ancillary data directory after fetch if it has not
    # appeared there in the meantime.  This will mean I fetch twice
    # when I could have waited, but at least I don't get started
    # processing with incompletely downloaded files and fail.
    # This does mean I need to make a similar patch where the GFS
    # grib2 files are converted to hdf - 
    my ($site, $file, $local) = (shift, shift, shift);
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $v = '';
    ($options{verbose} == 3) and $v = '-v';
    -e "$local/$file" and do {
      ($options{verbose} == 3) and print "$name\: $file FOUND in local repo\n";
      return 1;
    };
    ($options{quarantine}) and do {
      ($options{verbose} >= 1) and 
        warn "WARNING: $file NOT FOUND in local repo AND --quarantine";
      return 0;
    };
    open(my $fh, '>script_file.lftp') or die "$name\: Cannot open script_file.lftp $!";
    print $fh "set dns:order inet inet6\n";            # default is 'inet6 inet'
    #print $fh "set net:timeout 1\n";                   # default is '5'
    #print $fh "set net:max-retries 5\n";               # default is '10'
    #print $fh "set net:persist-retries 1\n";           # default is '5'
    #print $fh "set net:reconnect-interval-base 6\n";   # default is '60'
    print $fh "mirror $v -L -n --no-empty-dirs $site -I $file .\n";
    close($fh);
    my $cmd = "$LFTP -f script_file.lftp";
    system($cmd);
    -e $file or do {
      ($options{verbose} >= 2) and print "$name\: $file NOT FOUND in $site\n";
      return 0; };
    my $fullfile = File::Spec -> catfile($local, $file);
    -e $fullfile and return 1; # put there by another process while you were busy?
    $cmd = "mv $file $fullfile";
    system($cmd) == 0 or return 0;
    return 1;
  }

# --------------------
  sub make_aggCntlFile
# ====================
# make_aggCntlFile($cntlfile, \%options, @files);
  {
    # Make a control file for the supplied viirsAggregator
    # $cntlfile is the name of the control file
    # %options is passed for $options{input_dir} and $options{verbose}
    #   $options{input_dir} is the input directory of files to be aggregated
    #   $options{verbose} the verbosity level
    # @files are the files that comprise the swath
    my $cntlfile = shift;
    my %options = %{(shift)};
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($YYYY, $MM, $DD, $hh, $mm, $ss) = 'N/A' x 6;
    ($options{verbose} == 3) and print "$cntlfile disagg\n";
    my @geo = (); my $geo_type = 'GMODO';
    ($options{xfiles}) and $geo_type = 'GMTCO';
    foreach (@files) { (index($_, $geo_type) > -1) and push(@geo, $_); }
    ($#geo < 0) and die "$name\: Nothing to aggregate\nAbort";
    my $indir = 'disagg';
    my $start_geo = File::Spec -> catfile($indir, $geo[0]);
    my $stop_geo = File::Spec -> catfile($indir, $geo[$#geo]);
    my @start_datetime = (); my @stop_datetime = ();
    if ($options{xfiles}) {
      @start_datetime = get_datetime_from_gmtco($start_geo, 'Beginning');
      @stop_datetime = get_datetime_from_gmtco($stop_geo, 'Ending');
    } else {
      @start_datetime = get_datetime_from_gmodo($start_geo, 'Beginning');
      @stop_datetime = get_datetime_from_gmodo($stop_geo, 'Ending');
    }
    ($options{granule}) and do {
      $start_datetime[5] = 0;
      $start_datetime[4] = int(ceil($start_datetime[4]/10))*10;
      $stop_datetime[5] = 0;
      $stop_datetime[4] = int(floor($stop_datetime[4]/10))*10;
    };
    ($YYYY, $MM, $DD, $hh, $mm, $ss) = @start_datetime;
    my $start_gmtime = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
    ($YYYY, $MM, $DD, $hh, $mm, $ss) = @stop_datetime;
    my $stop_gmtime = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
    ($options{verbose} == 3) and print "time: $start_gmtime $stop_gmtime\n";
    my $granuleSize = sprintf("%02d", ($stop_gmtime - $start_gmtime)/60 + 1);
    ($options{verbose} == 3) and print "  granuleSize = $granuleSize\n";
    ($options{verbose} == 3) and print "start:  @start_datetime\n";
    ($options{verbose} == 3) and print "  end:  @stop_datetime\n";
    open (my $fh, ">$cntlfile") or die "$name\: Cannot open $cntlfile for writing\nAbort";
    print $fh "&cntlFile\n";
    printf $fh ("  dataStartTime = %04d, %02d, %02d, %02d, %02d, %02d\n", @start_datetime);
    printf $fh ("  dataStopTime  = %04d, %02d, %02d, %02d, %02d, %02d\n", @stop_datetime);
    print $fh "  granuleSize = $granuleSize\n";
    print $fh "  satID = npp\n";
    print $fh "  marginSize = 0\n";
    print $fh "  geoProdID = '$geo_type'\n";
    print $fh "  dataProdIDs = 'SVM05', 'SVM07', 'SVM10', 'SVM12', 'SVM15', 'SVM16'\n";
    print $fh "  geoInputDir = \'$indir\'\n";
    print $fh "  dataInputDirs = \'$indir\'\n";
    print $fh "    \'$indir\'\n" x 5;
    print $fh "  geoOutputDir = 'input'\n";
    print $fh "  dataOutputDirs = 'input'\n";
    print $fh "    'input'\n" x 5;
    print $fh "  compression = 'none'\n";
    print $fh "  uniformity = .T.\n";
    print $fh "  reportFlags = .T., .T.\n";
    print $fh "/\n";
    close ($fh) or die "$name\: Cannot close $cntlfile\nAbort";
    return;
  }

# ----------------------------
  sub make_avhrr_L1B_bigEndian
# ============================
# $outfile = make_avhrr_L1B_bigEndian($infile, $destdir);
  {
    # $infile is the name of the avhrr L1B file to test/convert endian
    # $outfile is the big endian output filename created (without path)
    # $destdir is where $outfile is to be stored, either an absolute path
    # or relative to the temporary working directory (which is $PWD).
    #
    # $outfile is extracted from the header of $infile, bytes 23-64 inclusive
    # (42 bytes). In practice, in run_acspo.pl, $destdir is set to 'input' so
    # that $outfile is created in WORK_DIR/input - this is to make sure that
    # we don't clobber data files in WORK_DIR if the user happens to specify
    # WORK_DIR as the place where the original L1B files are stored. For files
    # named like "hrpt_noaa19_20150210_1937_30966.l1b" no problem, but for
    # files named like "NSS.HRPT.NN.D06032.S0122.E0317.B0361819.WI", problem,
    # since this is the same as the filename stored in its header!
    #
    # NOTE: We coalesce repeated character variables to strings in the $Lhdr and
    # $Lrec buffer specification (because we are not actually applying any logic
    # to characters), but only after 'a42a8' in $Lhdr, since the a42 field is to
    # be used to define $outfile.
    #
    # I am only supporting $infile that is already in NOAA-packed format (but
    # which may be prepended with an optional 512 byte archive header) and
    # $infile that is in AAPP format. This function does not depend on filename
    # or endian to figure out what to do. It looks for a valid date to assess
    # endianess and specified record length to determine AAPP or NOAA HRPT.L1B
    # format. See NWPSAF-MF-UD-003_Formats.pdf for details. Note that rescaling
    # of avh_calvis and avh_calir are described in that document, but that the
    # re-ordering of avh_calir is not (it was deduced by inspection of AAPP and
    # NOAA HRPT.L1B files) as was division by 100 of avh_calir[(0,3,6,9,12,15)]
    # instead of the the documented 1000 (in NWPSAF-MF-UD-003_Formats.pdf).
    # The copying across from non-zero operational/test/prelaunch sets into
    # zero-valued operational/test/prelaunch sets is entirely my own invention
    # and I do this because I do not know which set ACSPO is using. In this
    # process I prefer operational over prelaunch over test.
    #
    # This function is peppered with "magic numbers" that are buffer lengths
    # and positions of important test values. My justification is that I think
    # it will make code maintenance easier, rather than abstracting out to
    # (even nicely descriptive) variable names.
    #
    use integer; # this allows me to do integer arithmetic and bit shifting
    my ($infile, $destdir) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $shortname = basename($infile);
    my @data = ();
    my ($Lhdr, $Bhdr, $Lrec, $Brec, $count, $buf) = undef;
    # Buffer specification for just the first 512 bytes of header
    $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv6a396'; # a396 fills out to 512 bytes
    $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g; # B is big-endian version
    # Open the input file ready for binary mode reading
    open my $fh, "<$infile" or die "$name\: open $shortname for read failed\nAbort";
    binmode $fh;
    # Find if $infile is big or little endian and with or without 512b header
    my ($skipHead, $Lendian, $Bendian) = (0, 0, 0);
    my ($avh_h_siteid, $avh_h_reclg, $avh_h_dataname) = ('XXX', 0, 'N/A');
    foreach (0..1) {
      $skipHead = $_;
      $count = read($fh, $buf, 512) or die "$name\: read $shortname failed\nAbort";
      ($count == 512) or die "$name\: Header count != reclen $shortname\nAbort";
      foreach my $hdr (($Lhdr, $Bhdr)) {
        @data = unpack($hdr, $buf);
        my ($avh_h_startdatayr, $avh_h_startdatady) = ($data[18], $data[19]);
        ($avh_h_startdatayr > 1960 and $avh_h_startdatayr < 2038) and do {
          ($avh_h_startdatady > 0 and $avh_h_startdatady < 367) and do {
            ($hdr eq $Lhdr) and $Lendian = 1;
            ($hdr eq $Bhdr) and $Bendian = 1;
            $avh_h_siteid = $data[0];
            $avh_h_reclg = $data[5];
            $avh_h_dataname = $data[11];
          };
        };
      }
      ($Lendian xor $Bendian) and last;
    }
    close $fh;
    # Check we have found a file that we know how to handle
    ($Lendian xor $Bendian) or die "$name\: $shortname endian indeterminate\nAbort";
    my $outfile = $avh_h_dataname;
    $outfile =~ s/^\s+//; $outfile =~ s/\s+$//; $outfile =~ s/ //g;
    (length($outfile) != 42) and die "$name\: L1B filename $outfile funky\nAbort";
    ($avh_h_reclg == 15872 or $avh_h_reclg == 22016) or 
      die "$name\: $shortname avh_h_reclg neither 15872 nor 22016\nAbort";
    my $fullfile = File::Spec -> catfile($destdir, $outfile);
    # Now we can define the complete header specification ...
    $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv6Vv2Vv14Va16v5v31V2V18a8v8V18v132';
    ($avh_h_reclg == 22016) and $Lhdr = $Lhdr.'v10664';
    ($avh_h_reclg == 15872) and $Lhdr = $Lhdr.'v7592';
    $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g; # Big-endian version
    # ... and the complete data record specification for reading ...
    ($avh_h_reclg == 22016) and $Lrec = 'v4Vv6V2v4V2V66V2v160V104v104v10240V2v35v30v67';
    ($avh_h_reclg == 15872) and $Lrec = 'v4Vv6V2v4V2V66V2v160V104v104V3414V2v35v30v67v340';
    $Brec = $Lrec; $Brec =~ s/v/n/g; $Brec =~ s/V/N/g; # Big-endian version
    # ... and for writing, always big-endian ...
    my $Fhdr = 'a3an9a42a8n4Nn2N2n2Nn6Nn2Nn14Na16n5n31N2N18a8n8N18n132n7592';
    my $Fout = 'n4Nn6N2n4N2N66N2n160N104n104N3414N2n35n30n67n340';
    # Re-open file for binary reading
    open $fh, "<$infile" or die "$name\: open $shortname for read failed\nAbort";
    binmode $fh;
    # Skip over prepended header?
    ($skipHead) and do {
      $count = read($fh, $buf, 512) or die "$name\: read $shortname failed\nAbort";
      ($count == 512) or die "$name\: Header count != reclen $shortname\nAbort"; };
    # Big endian and no header, just link to file ...
    ($Bendian and !$skipHead) and do {
      symlink($infile, $fullfile) or die "$name\: Cannot make link to $shortname\nAbort";
      return $outfile; }; # DONE
    # ... otherwise we must loop over read/write
    open my $FH, ">$fullfile" or die "$name\: open $fullfile for write failed\nAbort";
    binmode $FH;
    $count = read($fh, $buf, $avh_h_reclg) or die "$name\: read $shortname failed\nAbort";
    ($count == $avh_h_reclg) or die "$name\: Header count != avh_h_reclg $shortname\nAbort";
    ($Lendian) and @data = unpack($Lhdr, $buf);
    ($Bendian) and @data = unpack($Bhdr, $buf);
    $data[5] = 15872;
    $data[6] = 15872;
    $data[14] = 308;
    $buf = pack($Fhdr, @data[(0..7929)]);
    print $FH $buf or die "$name\: write $outfile failed\nAbort"; 
    my @avh_calvis_scale_base0_idx = (0,2,5,7,10,12,15,17,20,22,25,27,30,32,35,37,40,42);
    my @avh_calvis_ops_base0_idx = (0..4,15..19,30..34);
    my @avh_calvis_tst_base0_idx = (5..9,20..24,35..39);
    my @avh_calvis_pre_base0_idx = (10..14,25..29,40..44);
    my @avh_calir_scale_base0_idx = (0,3,6,9,12,15);
    my @avh_calir_swaps_base0_idx = (2,1,0,5,4,3,8,7,6,11,10,9,14,13,12,17,16,15);
    my @avh_calir_ops_base0_idx = (0..2,6..8,12..14);
    my @avh_calir_tst_base0_idx = (3..5,9..11,15..17);
    while (!eof($fh)) {
      $count = read($fh, $buf, $avh_h_reclg);
      ($count == $avh_h_reclg) or warn "$name\: WARNING Read count != avh_h_reclg $shortname\n";
      ($count == $avh_h_reclg) and do {
        ($avh_h_reclg == 22016 or $Lendian) and do {
          # we are going to have to unpack the buffer
          ($Lendian) and @data = unpack($Lrec, $buf);
          ($Bendian) and @data = unpack($Brec, $buf);
          @data = unpack($Lrec, $buf);
          ($avh_h_reclg == 22016) and do {
            # calvis modifications
            my @avh_calvis = @data[(19..63)];
            foreach (@avh_calvis_scale_base0_idx) {
              my $i = $_+1;
              $avh_calvis[$_] =  $avh_calvis[$_] / 1000;
              if ($avh_calvis[$i] > 0) {
                if ($avh_calvis[$i] < 2147483648) {
                  $avh_calvis[$i] = $avh_calvis[$i] / 10;
                } else {
                  $avh_calvis[$i] =  ($avh_calvis[$i] -  4294967296) / 10 ;
                }
              }
            }
            # Which set (ops, pre, tst) is used?  Dunno, so copy into empty ones
            my $max_ops = (sort { $b <=> $a } @avh_calvis[@avh_calvis_ops_base0_idx])[0];
            my $max_tst = (sort { $b <=> $a } @avh_calvis[@avh_calvis_tst_base0_idx])[0];
            my $max_pre = (sort { $b <=> $a } @avh_calvis[@avh_calvis_pre_base0_idx])[0];
            my @idx = @avh_calvis_pre_base0_idx;
            ($max_ops > 1) and @idx = @avh_calvis_ops_base0_idx;
            ($max_ops < 1) and @avh_calvis[@avh_calvis_ops_base0_idx] = @avh_calvis[@idx];
            ($max_pre < 1) and @avh_calvis[@avh_calvis_pre_base0_idx] = @avh_calvis[@idx];
            ($max_tst < 1) and @avh_calvis[@avh_calvis_tst_base0_idx] = @avh_calvis[@idx];
            @data[(19..63)] = @avh_calvis;
            # calir modifications, NWPSAF says 1000 but my inspection reveals 100 as factor
            my @avh_calir = @data[(64..81)];
            foreach (@avh_calir_scale_base0_idx) {
              $avh_calir[$_] =  $avh_calir[$_] / 100;
            }
            my @avh_calir_copy = @avh_calir;
            # This reversal is from my inspection, not in NWPSAF-MF-UD-003_Formats.pdf
            @avh_calir = @avh_calir_copy[@avh_calir_swaps_base0_idx];
            # Which set (ops, tst) is used?  Dunno, so copy into empty ones
            $max_ops = (sort { $b <=> $a } @avh_calir[@avh_calir_ops_base0_idx])[0];
            $max_tst = (sort { $b <=> $a } @avh_calir[@avh_calir_tst_base0_idx])[0];
            @idx = @avh_calir_tst_base0_idx;
            ($max_ops > 1) and @idx = @avh_calir_ops_base0_idx;
            ($max_ops < 1) and @avh_calir[@avh_calir_ops_base0_idx] = @avh_calir[@idx];
            ($max_tst < 1) and @avh_calir[@avh_calir_tst_base0_idx] = @avh_calir[@idx];
            @data[(64..81)] = @avh_calir;
            # NOAA says these are zero, so I will make them just that
            @data[(82..84)] = (0, 0, 0);
          }; 
        };
        # pack the avh_hrpt, 3x10-bit per 32-bit word 
        my @avh_hrpt_unpacked = @data[(455..10694)];
        my @avh_hrpt_packed = (0) x 3414;
        my $k = 0;
        for (my $j=0; $j<=$#avh_hrpt_unpacked; $j++) {
          my $i = (20,10,0)[$j % 3];
          $avh_hrpt_packed[$k] += ($avh_hrpt_unpacked[$j] << $i);
          ($i == 0) and $k++;
        }
        my @tail = (0) x 340;
        # construct an outut record vector ...
        my @out = (@data[(0..454)], @avh_hrpt_packed, @data[(10694..10828)], @tail);
        # ... and pack it into a raw buffer according to spec in $Fout
        $buf = pack($Fout, @out);
        # write the record
        print $FH $buf or die "$name\: write $outfile failed\nAbort";
      };
    }
    close $fh; close $FH;
    return $outfile;
  }

# -------------
  sub make_PCFs
# =============
# make_PCFs($tmpdir, \%sources, \%targets, %options);
  {
    # $tmpdir is fully-qualified pathname of temporary working directory.
    # \%sources are the files to use as PCF templates 
    # \%targets are the PCF files in $tmpdir derived from modifying \%sources
    # %options are the command line options, parsed and validated.
    # $pcf is fully-qualified PCF filepath for this run.
    # This function bails within on failure. Note that I left the source files
    # in their original form (on development machine thunder) and just swapped
    # the correct values for a CSPP_ACSPO installation.
    my $tmpdir = shift;
    my %sources = %{(shift)};
    my %targets = %{(shift)};
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $static_anc = File::Spec -> catfile($HOME, 'run', 'static_anc');
    foreach my $key (keys %sources) {
      my $tmp = File::Spec -> catfile($options{run_dir}, $sources{$key});
        (grep {$options{sat} eq $_} qw/npp aqua terra/) and do {
          ($key eq 'aer') and next; };
      my $pcf = $targets{$key};
      open(my $fhI, "<$tmp") or die "$name\: Cannot open $tmp for read.\nAbort";
      open(my $fhO, ">$pcf") or die "$name\: Cannot open $pcf for write.\nAbort";
      while (<$fhI>) { chomp;
        $_ =~ s|dir=../output/|dir=$options{work}/|g;
        $_ =~ s|../static_anc|$static_anc|g;
        my @part = split(/\=/, $_);
        my $keyword = $part[0];
        /dynamic_anc/ and $_ = "$keyword=$tmpdir";
        /input_dir=aggregate/ and $_ = "$keyword=input/";
        /input_dir=..\/input\// and $_ = "$keyword=input/";
        ($options{ghrsst} eq '0.1') and do {
          $_ =~ s|CMC0.2deg|CMC0.1deg|g;
          $_ =~ s|GLOB-v02.0-fv02.0|GLOB-v02.0-fv03.0|g; };
        print $fhO "$_\n";
      }
      close($fhI) or die "$name\: Cannot close file $tmp.\nAbort";
      close($fhO) or die "$name\: Cannot close file $pcf.\nAbort";
    }
    my $list = '1bFilesList.txt';
    open(my $fhO, ">$list") or die "$name\: Cannot make $list.\nAbort";
    foreach (@{$options{process_files}}) {
      print $fhO "$_\n";
    }
    close($fhO) or die "$name\: Cannot close file $list.\nAbort";
    return;
  }

# --------------------------
  sub npp_granules_to_swaths
# ==========================
# %swaths = npp_granules_to_swaths(@files);
  {
    # Break list of files into groups that comprise contiguous overpass
    # swaths ready for the NOAA-provided viirsAggregator to concatenate.
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $time_window = 60 * $AGG_WINDOW_MINUTES;
    my %swaths = ();
    my ($YYYY, $MM, $DD, $hh, $mm, $found) = undef;
    foreach (@files) {
      $YYYY = substr($_, 11, 4);
      $MM = substr($_, 15, 2);
      $DD = substr($_, 17, 2);
      $hh = substr($_, 21, 2);
      $mm = substr($_, 23, 2);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm) or
        die "$name\: Date from npp file $_ implausible\nAbort";
      my $t = timegm_nocheck(0, $mm, $hh, $DD, $MM-1, $YYYY);
      $found = 0;
      foreach my $key (sort keys %swaths) {
        (abs($key - $t) < $time_window) and do {
          push(@{$swaths{$key}}, $_);
          $found = 1; last; };
      }
      ($found) or @{$swaths{$t}} = ($_);
    }
    return %swaths;
  }

# ---------------------------
  sub pre_fetch_generic_acspo
# ===========================
# ($url, $mask, $local)  = pre_fetch_generic_acspo($code, @datetime, %dynanc_loc);
#  ...which could be ... = pre_fetch_generic_acspo('GFS', @gfs_datetime, %dynanc_loc);
#  ...    or, say,   ... = pre_fetch_generic_acspo('CMC', @cmc_datetime, %dynanc_loc);
#  where @datetime = ($YYYY, $MM, $DD, $ddd, $hh);
  {
    # Spawned from pre_fetch_generic() in the IDEA-I project. I have dumped the
    # authentication, which I do not need to access SSEC servers, and changed the
    # keywords to include a protocol since some data is under http://, not ftp://,
    # but lftp can still be used to fetch (just got to construct URL differently).
    # Determines the $url for lftp (without authentication), the $mask
    # to use for downloads and the name of the local directory that will
    # accommodate the downloaded data, $local. The date and time information
    # ($YYYY, $MM, $DD, $ddd, $hh) are for the beginning of forecast data required
    my $code = shift;
    my ($YYYY, $MM, $DD, $ddd, $hh) = (shift, shift, shift, shift, shift);
    my (%dynanc_loc) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $YY = substr($YYYY, 2 , 2);
    my $proto = $dynanc_loc{$code.'PROTO'};
    my $site = $dynanc_loc{$code.'SITE'};
    my $remote = $dynanc_loc{$code.'REMOTE'};
    my $mask = $dynanc_loc{$code.'TEMPLATE'};
    my $local = $dynanc_loc{$code.'LOCAL'};
    my @tokens = qw/%Y %y %m %d %j %H/;
    my @values = ($YYYY, $YY, $MM, $DD, $ddd, $hh);
    for (my $j=0; $j<=$#tokens; $j++) {
      $local =~ s/$tokens[$j]/$values[$j]/g;
      $remote =~ s/$tokens[$j]/$values[$j]/g; 
      $mask =~ s/$tokens[$j]/$values[$j]/g; 
    }
    # ********** MAKE PATH SO THAT WARNINGS CAN BE REPORTED **********
    if (! -d $local) {
      eval { mkpath($local) };
      if ($@) { warn "$name\: WARNING $local cannot be made\n"; }
    }
    my $url = "$proto$site/$remote";
    return ($url, $mask, $local);
  }

# ---------------
  sub pretty_dump
# ===============
# pretty_dump($title, %hash);
  {
    # Do a "pretty print" of %hash by effectively grepping Dumper().
    # Assumes standard module Data::Dumper has been loaded.
    my $title = 'DEBUG '.shift;
    my (%hash) = @_;
    print '-' x length($title)."\n";
    print "$title\n";
    print '=' x length($title)."\n";
    foreach (split('\n', Dumper(\%hash))) {
      chomp;
      ((/VAR1/) or (/\}\;/)) and next;
      $_ =~ s/  //g; # shorten up the whitespace
      print "$_\n";
    }
    print "$title _END_\n";
  }

# ---------------
  sub token_value
# ===============
# $value = token_value($token, $file, $mask);  ...example follows...
# $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
  {
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                                                                           #
    #   WARNING: This function not used in ACSPO but kept as token reference    #
    #                                                                           #
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #
    # Takes the filename $file (within which tokens have already been expanded),
    # and the $mask (within which tokens are unexpanded) and, for a given $token
    # return the $value extracted from $file. Here is an example ...
    #   $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
    # which would return $value as '332'.
    # These are the supported tokens, returns the first occurance if more than one:
    #   %Y  Year with century. (1582-9999)
    #   %y  Year without century (00–99).
    #   %m  Month as decimal number (01–12).
    #   %j  Day of year as decimal number (001–366).
    #   %d  Day of the month as decimal number (01–31).
    #   %H  Hours as decimal number (00–23).
    #   %M  Minute as decimal number (00–59).
    #   %S  Second as decimal number (00–59).
    #   %F  3-digit forecast hour (000-999).
    #   %f  2-digit forecast hour (00-99).
    my $token = shift;
    my $file = shift;
    my $mask = shift;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $value = undef;
    my $expand = $mask;
    my @tokens = qw/  %Y %y %m  %j %d %H %S %M  %F %f/;
    my @widths =   (   4, 2, 2,  3, 2, 2, 2, 2,  3, 2);
    my $width = 0;
    for (my $j=0; $j<=$#tokens; $j++) {
      ($token eq $tokens[$j]) and do { $width = $widths[$j]; next; };
      my $blank = 'X' x $widths[$j];
      $expand =~ s/$tokens[$j]/$blank/g; 
    }
    my $i = index($expand, $token);
    if ($i > -1) {
      $value = substr($file, $i, $width);
    }
    return $value;
  }

# ---------------
  sub trim_string
# ===============
# $new_string = trim_string($old_string);
  {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
  }

# ---------
  sub usage
# =========
# usage();
  {
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    open(my $fh, "<$0");
    while (<$fh>) { chomp;
      if ( /START-HELP/ ) {
        while (<$fh>) { chomp;
          ( /END-HELP/ ) and do { close($fh); exit(9); };
          $_ =~ s/\$CSPP_ACSPO_HOME/$ENV{CSPP_ACSPO_HOME}/g;
          my $help = (split(/\#/, $_))[1]; # remove leading hash
          $help and print "$help\n" or print "\n";
        }
      }
    }
    die "$name\: ERROR! Have you tinkered with help in script header?\nAbort";
  }

# --------------------
  sub validate_options
# ====================
# %options = validate_options($filemask, $satdir, %options);
  {
    # $filemask is for globbing the nominated satellite data.
    # $satdir is the path component to construct build and run directories
    # %options are the command line options to "validate".
    # Fatal errors in here terminate with die without showing usage() message.
    # The most important task is to convert $options{input_files} to a scalar
    # $options{input_dir} and filename list put back in $options{input_files}.
    # NOTE: The directory for dynamic ancillary data, $options{dynanc}, is not
    # made because it could contain unexpanded date/time tokens, and I do not
    # know how to expnad those tokens until the input data files are examined
    # for their data's date and time - NOT the job of this function.
    my ($filemask, $satdir) = (shift, shift);
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $exe_mask = 'ACSPO-UNKNOWN';
    ($options{sat} eq 'npp') and $exe_mask = 'ACSPO-V_V?.??';
    (grep {$_ eq $options{sat}} qw/aqua terra/) and $exe_mask = 'ACSPO-M_V?.??';
    (grep {$_ eq $options{sat}} qw/metopA metopB/) and $exe_mask = 'ACSPO-A_V?.??';
    (grep {$_ eq $options{sat}} qw/n15 n16 n17 n18 n19/) and $exe_mask = 'ACSPO-A_V?.??';
    my $exe_glob = File::Spec -> catfile($HOME, 'bin', $exe_mask);
    $options{exe} = (reverse(sort(glob($exe_glob))))[0];
    (-e $options{exe}) or die "$name\: program $options{exe} not found.\nAbort";
    (-x $options{exe}) or die "$name\: program $options{exe} not executable.\nAbort";
    (grep {$_ eq $options{ghrsst}} qw/0.1 0.2 by_date/) or
      die "$name\: --ghrsst=$options{ghrsst} is not valid.\nAbort";
    $options{destripe} = 'N/A';
    ($options{sat} eq 'npp') and 
      $options{destripe} = File::Spec -> catfile($HOME, 'bin', 'destripe_viirs');
    (grep {$_ eq $options{sat}} qw/aqua terra/) and
      $options{destripe} = File::Spec -> catfile($HOME, 'bin', 'destripe_modis');
    ($options{destripe} eq 'N/A') or do {  
      (-e $options{destripe}) or die "$name\: program $options{destripe} not found.\nAbort";
      (-x $options{destripe}) or die "$name\: program $options{destripe} not executable.\nAbort"; };
    $options{run_dir} = File::Spec -> catfile($HOME, 'run', $satdir);
    my $globber = $options{input_files};
    my $mask = 'N/A';
    my $altglob = $globber;
    (-d $options{input_files}) and do {
      my $base = $options{input_files};
      $globber = File::Spec -> catfile($base, $filemask);
      ($options{xfiles}) or do { # support a second DB filename format
        ($options{sat} eq 'aqua') and $mask = 'A*.{GEO,L1B_LAC}';
        ($options{sat} eq 'terra') and $mask = 'T*.{GEO,L1B_LAC}';
        $altglob = File::Spec -> catfile($base, $mask); }; };
    my @input_files = sort(glob($globber));
    ($#input_files < 0 and $mask ne 'N/A') and @input_files = sort(glob($altglob));
    ($#input_files < 0) and die "$name\: INPUTFILES not found.\nAbort";
    $options{input_dir} = dirname(abs_path($input_files[0]));
    $options{input_files} = ();
    foreach (@input_files) { push(@{$options{input_files}}, basename($_)); }
    foreach (qw/bias dynanc work/) {
      (-d $options{$_}) or do {
        ($options{verbose} >=2) and print "Making $_ DIR $options{$_}\n";
        mkpath($options{$_}) or
          die "$name\: $_ DIR $options{$_} cannot be made.\nAbort"; };
      (-w $options{$_}) or
        die "$name\: $_ DIR $options{$_} not writable.\nAbort";
    }
    return %options;
  }
